# /usr/bin/env python
#
# Handle some internal functions through HTTP server, and rest API
#
# Why ?
# Because this is far more powerful and beautiful than a telnet or
# a TCP socket
#
# Guillaume Penaud (gpenaud@easter-eggs.com)
# 05/2015

import json
from urllib.parse import urlparse
import logging

from http.server import BaseHTTPRequestHandler, HTTPServer


class ApiManager():

    def __init__(self, server):
        self.server = server
        self.config = server.config
        self.logger = server.logger

        # handle socket commands
        self.api_commands = {
            "load": self.handle_load_command,
            "dump": self.handle_dump_command,
            "flush": self.handle_flush_command,
            "get_limits": self.handle_get_limits_command,
            "get_periods": self.handle_get_periods_command,
            "reload_conf": self.handle_reload_configuration_command,
        }


# Server methods
# --------------------------------------------------------------------------

    def start(self):
        self.logger.debug("Start API server")

        hostname = self.config.get("api", {}).get("api_host", "0.0.0.0")
        port = self.config.get("api", {}).get("api_port", 9002)

        try:
            self.api_server = HTTPServer((hostname, port), ApiServer)
            
            ApiServer.set_api_commands(self.api_commands)
            ApiServer.set_logger(self.logger)

            try:
                self.api_server.serve_forever()
            # FIXME handle exception here ?
            except KeyboardInterrupt:
                pass
        except Exception as e:
            print(e.message)
            self.stop()

    def stop(self):
        self.logger.debug("Stop API server")

        try:
            self.api_server.server_close()
        except AttributeError as e:
            self.logger.debug("API server already stopped")
            return True
        
        # will apply at next serve_forever() iteration
        self.api_server.stopped = True


# API methods
# --------------------------------------------------------------------------
    
    # load cache file content to RAM memory
    def handle_load_command(self, args):
        return self.server.datamanager.load()

    def handle_dump_command(self, args):
        # Those are already up-to-date data
        return self.server.datamanager.dump()

    def handle_flush_command(self, args):
        return self.server.datamanager.flush()

    def handle_reload_configuration_command(self, args):
        return self.server.load_config()

    def handle_get_periods_command(self, args):
        return self.server.periods.keys()

    def handle_get_limits_command(self, args):
        user = args.get("user")
        limits = {}

        if user is None:
            raise Exception("Unable to retrieve limit, missing user")

        for period in self.handle_get_periods_command(args):
            limits.setdefault(period, self.server.ldapmanager.get_limit_from_ldap(user, period))

        return limits


# Api server
# --------------------------------------------------------------------------

class ApiServer(BaseHTTPRequestHandler):

    logger = None
    stopped = False

    def serve_forever(self):
        while self.stopped is False:
            self.handle_request()

    def parse_http_request(self, query_string):
        self.logger.debug("Parse API request")
        options = query_string.split('&')

        if len(options) == 0:
            raise Exception("No argument provided")

        parsed_options = [option.split('=') for option in options]

        return dict(parsed_options)

    def execute_command(self, arguments):
        self.logger.debug("Execute API command")
        try:
            if "cmd" not in arguments:
                raise Exception("Missing command parameter")

            command = arguments["cmd"]
            del arguments["cmd"]

            if command not in ApiServer.api_commands:
                raise Exception("Command not found")

            self.logger.info("Request command: [%s] with arguments: %s" % (command, arguments))
            result = ApiServer.api_commands[command](arguments)
        except Exception as e:
            self.logger.error("Error during API request: %s" % e)
            result = json.dumps({"status": "[ERROR] %s" % e.message})

        self.send_response(200)
        self.send_header("Content-type", "text/json")
        self.end_headers()

        self.wfile.write(b"%s \n" % result.encode("utf-8"))

    def do_GET(self):
        self.logger.debug("Handle GET API request")
        # request parsing
        parsed_path = urlparse(self.path)
        arguments = self.parse_http_request(parsed_path[4])

        # command execution
        return self.execute_command(arguments)

    def do_POST(self):
        self.logger.debug("Handle POST API request")
        # request parsing
        content_length = int(self.headers.getheader('content-length', 0))
        post_content = self.rfile.read(content_length)
        arguments = self.parse_http_request(post_content)

        # command execution
        return self.execute_command(arguments)

    @classmethod
    def set_logger(self, logger):
        self.logger = logger

    @classmethod
    def set_api_commands(self, commands):
        self.api_commands = commands
